/*****************************************************************************

This source code is made available to you under the terms of the Open
Development License.  This Package should be accompanied by a file that
contains the associated licensing terms.  If you are unable to find this file,
please refer to its location on the Internet at www.rocklyte.com/license.html.

******************************************************************************

###CLASS###
Name:      Variable
Version:   1.0
Status:    Stable
Date:      November 2002
ID:        ID_VARIABLE
Category:  Command
Author:    Rocklyte Systems
Copyright: Rocklyte Systems (c) 2000-2002.  All rights reserved.
Keywords:  variable, var, datatype, data, value, field

###DESCRIPTION###
<p>The Variable class is used to create, store and retrieve values during the
execution of a script.  Essentially it allows you to store any type of data
that you like using object compatible methods.  It also has its uses in
saving program options, desktop settings and other bits of information
that may need to be recalled at a later stage.</p>

<p>Data storage and retrieval is achieved through the use of unlisted fields,
so a variable object named 'storage' with fields 'x', 'y' and 'z' would be
created as follows when using DML:</p>

<pre>
   &lt;variable name="storage" &amp;x="1" &amp;y="2" &amp;z="3"/&gt;
</pre>

<p>For occasions when data needs to be stored permanently, the Variable class
supports file locations for data storage and retrieval.  Using our previous
example, we can store the x, y and z values in a configuration file, as
follows:</p>

<pre>
   [data]
   x = 1
   y = 2
   z = 3
</pre>

<p>To load the data file into a variable object, we would use this DML code:</p>

<pre>
   &lt;variable name="storage" location="data.cfg"/&gt;
</pre>

<p>When using a file location, bear in mind that the Variable class will
save all data back to the file when the Variable object is destroyed.</p>

###END###

*****************************************************************************/

#include <pandora/system/all.h>
#include <pandora/data/config.h>
#include <pandora/files/file.h>

static ERROR CMDInit(OBJECTPTR, struct KernelBase *);
static ERROR CMDExpunge(void);

#define VER_VARIABLE 1.0

EXPORT struct ModHeader ModHeader = {
   MODULE_HEADER_V1, CMDInit, NULL, NULL, CMDExpunge,
   JMP_DEFAULT, 0, CPU_PC, VER_VARIABLE, VER_KERNEL,
   "Rocklyte Systems",
   "Rocklyte Systems (c) 2000-2002.  All rights reserved.",
   "November 2002",
   "Variable"
};

static struct KernelBase *KernelBase = 0;
static OBJECTPTR VariableClass = 0;

/*****************************************************************************
** Class definition.
*/

struct VStore {
   OBJECT_HEADER       /* Class header */
   LONG Flags;         /* Flags */
   OBJECTPTR Config;

   PRIVATE_FIELDS
   OBJECTID ConfigID;
};

static ERROR GET_Location(struct VStore *, STRING *);
static ERROR GET_TotalSections(struct VStore *, LONG *);

static ERROR SET_Location(struct VStore *, STRING);

#define VF_NOREPLACE  0x00000001
#define VF_NOSAVE     0x00000002

#define VF_LOADCONFIG 0x80000000

static struct FieldDef VariableFlags[] = {
   { "NOREPLACE", VF_NOREPLACE },
   { "NOSAVE",    VF_NOSAVE },
   { NULL, NULL }
};

static struct FieldArray VariableFields[] = {
   { "Flags",    0, FDF_LONGFLAGS|FDF_RW, (LONG)&VariableFlags, NULL, NULL },
   //{ "Config",   0, FDF_CHILD|FDF_R,      ID_CONFIG, NULL, NULL },
   /*** Virtual Fields ***/
   { "Location",      0, FDF_STRING|FDF_RW, 0, GET_Location, SET_Location },
   { "Src",           0, FDF_STRING|FDF_RW, 0, GET_Location, SET_Location },
   { "TotalSections", 0, FDF_LONG|FDF_R,    0, GET_TotalSections, NULL },
   END_FIELD
};

struct mtStripSlash {
   STRING Variable;
};

static ERROR VAR_AccessObject(struct VStore *, APTR);
static ERROR VAR_Free(struct VStore *, APTR);
static ERROR VAR_GetUnlistedField(struct VStore *, struct acGetUnlistedField *);
static ERROR VAR_Init(struct VStore *, APTR);
static ERROR VAR_NewObject(struct VStore *, APTR);
static ERROR VAR_ReleaseObject(struct VStore *, APTR);
static ERROR VAR_StripSlash(struct VStore *, struct mtStripSlash *);
static ERROR VAR_SetUnlistedField(struct VStore *, struct acSetUnlistedField *);

static struct ActionArray VariableActions[] = {
   { AC_AccessObject,     VAR_AccessObject },
   { AC_Free,             VAR_Free },
   { AC_GetUnlistedField, VAR_GetUnlistedField },
   { AC_Init,             VAR_Init },
   { AC_NewObject,        VAR_NewObject },
   { AC_ReleaseObject,    VAR_ReleaseObject },
   { AC_SetUnlistedField, VAR_SetUnlistedField },
   { NULL, NULL }
};

/*****************************************************************************
** Method definitions
*/

#define MT_StripSlash -1

static struct FunctionField argsStripSlash[] = { {"Variable", ARG_STRING }, { NULL, NULL } };

static struct MethodArray VariableMethods[] = {
   { MT_StripSlash, VAR_StripSlash, "StripSlash", argsStripSlash, sizeof(struct mtStripSlash) },
   /* StripCharacters */
   /* Trim */
   /* Insert */
   /* Replace */
   /* UpperCase */
   /* LowerCase */
   /* Capitalise */
   { NULL, NULL, NULL, NULL, NULL }
};

/*****************************************************************************
** User Instructions.
*/

static BYTE VariableUsage[] = {
"The Variable command is designed for use in object scripts only;\n\
i.e. it will not work from the Command Line Interface.  Please\n\
refer to the Variable manual in the Pandora Engine SDK for more\n\
information on this command.\n"
};

/*****************************************************************************
** Command: Init()
*/

static ERROR CMDInit(OBJECTPTR argModule, struct KernelBase *argKernelBase)
{
   KernelBase = argKernelBase;

   return(CreateObject(ID_CLASS, NULL, (OBJECTPTR *)&VariableClass, NULL,
      FID_BaseClassID|TLONG, ID_VARIABLE,
      FID_Version|TFLOAT,    VER_VARIABLE,
      FID_Name|TSTRING,      "Variable",
      FID_Category|TLONG,    CCF_COMMAND,
      FID_Usage|TPTR,        VariableUsage,
      FID_Actions|TPTR,      VariableActions,
      FID_Methods|TPTR,      VariableMethods,
      FID_Fields|TPTR,       VariableFields,
      FID_Size|TLONG,        sizeof(struct VStore),
      TAGEND));
}

/*****************************************************************************
** Command: Expunge()
*/

static ERROR CMDExpunge(void)
{
   if (VariableClass) { Action(AC_Free, VariableClass, NULL); VariableClass = NULL; }
   return(ERR_Okay);
}

/*****************************************************************************
** Variable: AccessObject
*/

static ERROR VAR_AccessObject(struct VStore *Self, APTR Void)
{
   ERROR error;

   if (Self->ConfigID) {
      if ((error = AccessObject(Self->ConfigID, 0, &Self->Config)) != ERR_Okay) {
         DPrintF("@AccessObject()","[Variable:%d] Failed to access internal Config object #%d, error %d.", Self->Head.UniqueID, Self->ConfigID, error);
         Self->ConfigID = NULL; /* Someone has killed our configuration object */
      }
   }
   return(ERR_Okay);
}

/*****************************************************************************
** Variable: Free
*/

static ERROR VAR_Free(struct VStore *Self, APTR Void)
{
   struct acSaveToObject saveargs;
   STRING Location;
   OBJECTPTR File;
   LONG Flags;

   if (Self->ConfigID) {

      /* If a location has been used, save the details
      ** permanently before freeing the file.
      */

      if ((Self->Flags & VF_LOADCONFIG) AND (!(Self->Flags & VF_NOSAVE))) {
         if (!Self->Config) {
            AccessObject(Self->ConfigID, 0, &Self->Config);
         }

         if (Self->Config) {
            if (NewPrivateObject(ID_FILE, NULL, &File) IS ERR_Okay) {
               Flags = FL_NEW|FL_WRITE;
               SetField(File, FID_Flags, FT_LONG, &Flags);

               if (GetField(Self->Config, FID_Location, FT_POINTER, &Location) IS ERR_Okay) {
                  SetField(File, FID_Location, FT_POINTER, Location);

                  if (Action(AC_Init, File, NULL) IS ERR_Okay) {
                     saveargs.DestinationID = File->UniqueID;
                     saveargs.ClassID = NULL;
                     Action(AC_SaveToObject, Self->Config, &saveargs);
                  }
               }
               else DPrintF("@Free:","[Variable] Failed to acquire a Location from the Config object.");

               Action(AC_Free, File, NULL);
            }
            else DPrintF("@Free:","[Variable] Failed to create a File object for saving.");
         }
      }

      /*** Free the Config object ***/

      if (Self->Config) {
         acFree(Self->Config);
         ReleaseObject(Self->Config);
         Self->Config = NULL;
      }
      else ActionMsg(AC_Free, Self->ConfigID, NULL);
      Self->ConfigID = NULL;
   }

   return(ERR_Okay);
}

/****************************************************************************

###ACTION###
Name:  GetUnlistedField
Short: Retrieves stored field values from a variable object.
###END###

****************************************************************************/

static ERROR VAR_GetUnlistedField(struct VStore *Self, struct acGetUnlistedField *Args)
{
   return(Action(AC_GetUnlistedField, Self->Config, Args));
}

/*****************************************************************************
** Variable: Init
*/

static ERROR VAR_Init(struct VStore *Self, APTR Void)
{
   Action(AC_Init, Self->Config, NULL);
/*
   if (Self->Flags & VF_LOADCONFIG) {
      return(Action(AC_Init, Self->Config, NULL));
   }
*/
   return(ERR_Okay);
}

/*****************************************************************************
** Variable: NewObject
*/

static ERROR VAR_NewObject(struct VStore *Self, APTR Void)
{
   return(NewObject(ID_CONFIG, Self->Head.ObjectFlags|NF_CHILD, &Self->Config, &Self->ConfigID));
}

/*****************************************************************************
** Variable: ReleaseObject
*/

static ERROR VAR_ReleaseObject(struct VStore *Self, APTR Void)
{
   if (Self->Config) { ReleaseObject(Self->Config); Self->Config = NULL; }
   return(ERR_Okay);
}

/*****************************************************************************

###ACTION###
Name: SetUnlistedField
Short: Stores new field values in a variable object.
###END###

*****************************************************************************/

static ERROR VAR_SetUnlistedField(struct VStore *Self, struct acSetUnlistedField *Args)
{
   struct ConfigEntry *entries;
   LONG amtentries, i;

   if ((!Args) OR (!Args->Field) OR (!Args->Field[0])) {
      return(ObjectError(Self, ERH_SetField, ERR_Args));
   }

   /*** if NOREPLACE is set, check if the variable already exists ***/

   if (Self->Flags & VF_NOREPLACE) {
      if (GetFields(Self->Config, FID_Entries|TPTR,     &entries,
                                  FID_AmtEntries|TLONG, &amtentries,
                                  TAGEND) IS ERR_Okay) {

         for (i=0; i < amtentries; i++) {
            if (StrCompare(entries[i].Item, Args->Field, NULL, STR_MATCHLENGTH) IS ERR_Okay) {
               return(ERR_Okay);
            }
         }
      }
   }

   return(Action(AC_SetUnlistedField, Self->Config, Args));
}

/*****************************************************************************

###METHOD###
Name:  StripSlash
Short: Strips trailing forward and back slashes from variable fields.

###DESCRIPTION###
<p>This method will strip any given variable field of any back or forward
slashes that happen to be trailing at the end of the field string.  For
instance, using this method to strip a variable containing "athene:documents/"
will result in the string being reduced to "athene:documents".</p>

###INPUT###
STRING/Variable: The name of the variable field that you want to strip.

###RESULT###
ERR_Okay: All back and forward slashes were removed successfully.
ERR_Args: Invalid arguments were specified.
ERR_Search: The referenced variable name does not exist.

###END###

*****************************************************************************/

static ERROR VAR_StripSlash(struct VStore *Self, struct mtStripSlash *Args)
{
   #undef SIZE_BUFFER
   #define SIZE_BUFFER 500
   struct acGetUnlistedField get;
   struct mtAddConfigEntry add;
   UBYTE buffer[SIZE_BUFFER];
   LONG i;

   if ((!Args) OR (!Args->Variable)) return(ObjectError(Self, MT_StripSlash, ERR_Args));

   get.Field  = Args->Variable;
   get.Buffer = buffer;
   get.Size   = SIZE_BUFFER;
   if (Action(AC_GetUnlistedField, Self, &get) IS ERR_Okay) {
      for (i=0; buffer[i]; i++);
      if ((buffer[i-1] IS '/') OR (buffer[i-1] IS '\\')) {
         buffer[i-1] = 0;

         add.Section = "Variables";
         add.Item    = Args->Variable;
         add.Data    = buffer;
         return(Action(MT_AddConfigEntry, Self->Config, &add));
      }
      return(ERR_Okay);
   }
   else return(ERR_Search);
}

/*****************************************************************************

###FIELD###
Name:   Flags
Short:  Special flag settings.
Type:   LONG
Prefix: VF_
Status: Read/Write

###DESCRIPTION###
<p>This field contains flags that affect object behaviour. Currently available
flags are:</p>

<blockquote>
 <p><u>NOREPLACE</u><br>
 The NoReplace flag is useful when the over-writing of variable data is
 undesirable.  Setting this field will still allow new variables to be stored
 in the object, but you won't be able to update the values.</p>
</blockquote>

###END###

*****************************************************************************/

/*****************************************************************************

###FIELD###
Name:   Location
Short:  Variables can be loaded from a configuration file if you set this field.
Type:   STRING
Status: Get/Set

###DESCRIPTION###
<p>Setting the location field tells the Variable object to load its variables
from a file location.  The file must be written in the standard configuration
format (refer to the Config class for details).  When the Variable class
is destroyed, all current field stores will be saved back to the file
location so that any changes are retained.</p>

###SEE ALSO###
Class: Config

###END###

*****************************************************************************/

static ERROR GET_Location(struct VStore *Self, STRING *Value)
{
   if (!Self->Config) return(ERR_Failed);
   return(GetField(Self->Config, FID_Location, FT_POINTER, Value));
}

static ERROR SET_Location(struct VStore *Self, STRING Value)
{
   Self->Flags |= VF_LOADCONFIG;
   return(SetField(Self->Config, FID_Location, FT_POINTER, Value));
}

/*****************************************************************************

*****************************************************************************/

static ERROR GET_TotalSections(struct VStore *Self, LONG *Value)
{
   return(GetField(Self->Config, FID_TotalSections, FT_LONG, Value));
}
