/*****************************************************************************

This source code is made available to you under the terms of the Open
Development License.  This Package should be accompanied by a file that
contains the associated licensing terms.  If you are unable to find this file,
please refer to its location on the Internet at www.rocklyte.com/license.html.

******************************************************************************

###CLASS###
Name:      Button
Version:   1.0
ID:        ID_BUTTON
Status:    Beta
Category:  GUI
Date:      November 2003
Author:    Rocklyte Systems
Copyright: Rocklyte Systems, 2003.  All rights reserved.
Keywords:  button, click, press, box
Short:     The Button class manages the display and interactivity of user buttons.

###DESCRIPTION###
<p>The Button class simplifies the creation and management of buttons as
part of the user interface.  New buttons are typically created by declaring
the graphical dimensions and the text to be displayed within them.  The Button
class allows for the specifics of the button to be altered, such as the colours
and the font style.</p>

<p>The definitions for new buttons are loaded by default from the
environment file "templates:buttondef.xml".  You can change the template file
prior to initialisation by setting the Template field.  Note that any values
set by the template will override the existing field settings in the
button.</p>

<p>You will need to configure the button so that when it is clicked, it
performs an action.  There are three acceptable methods that you can use to
achieve this.  The first is to set the <a href="#ActionScript">ActionScript</a>
field so that it refers to a script that executes when the button is pressed.
The second is to initialise child objects to the button and they will be
executed on activation. The third option is to listen to the Activate action
by calling the <a href="../modules/kernel.html#SubscribeAction()">SubscribeAction()</a>
function on the button.</p>

<p>The following example illustrates how you might create a cancel button in
DML.  Notice that the Action object is initialised as static and will be
executed when the button is clicked:</p>

<pre>
  &lt;button xoffset="[window.leftmargin]" yoffset="[window.bottommargin]"
    text="Cancel" width="70"&gt;
    &lt;action static call="free" object="[window]"/&gt;
  &lt;/button&gt;
</pre>

###END###

*****************************************************************************/

#include <pandora/system/all.h>
#include <pandora/graphics/all.h>
#include <pandora/files/all.h>
#include <pandora/misc/all.h>
#include <pandora/modules/strings.h>
#include <pandora/modules/filesystem.h>
#include <pandora/input/joyport.h>
#include <pandora/input/keyboard.h>
#include <pandora/data/xml.h>
#include <pandora/functions/set.c>

static ERROR CMDInit(OBJECTPTR, struct KernelBase *);
static ERROR CMDExpunge(void);

#define VER_BUTTON  1.0

MODULE_HEADER = {
   MODULE_HEADER_V1, CMDInit, NULL, NULL, CMDExpunge,
   JMP_DEFAULT, 0, CPU_PC, VER_BUTTON, VER_KERNEL,
   "Rocklyte Systems",
   "Rocklyte Systems (c) 2003. All rights reserved.",
   "November 2003",
   "Button"
};

static struct KernelBase  *KernelBase;
static struct StringsBase *StringsBase;
static struct FileSystemBase *FileSystemBase;
static OBJECTPTR ButtonClass = NULL;
static OBJECTPTR modStrings = NULL;
static OBJECTPTR modFileSystem = NULL;

enum {
   STATE_ENTERED=1,
   STATE_EXITED,
   STATE_INSIDE
};

/*****************************************************************************
** Object structure.
*/

struct Button {
   OBJECT_HEADER              /* Class header */
   OBJECTID RegionID;         /* Drawable region created by the button object */
   OBJECTID DrawableID;       /* The drawable target for the button graphic */
   OBJECTID InputID;          /* Keyboard object to use for receiving input */
   LONG     Flags;            /* Special options */
   struct   Font *Font;       /* Font control object */
   LONG     EnterFrame;       /* Frame to display when the mouse enters the button area */
   LONG     ExitFrame;        /* Frame to revert to when the mouse leaves the button area */
   LONG     ClickFrame;       /* Frame to display when the button is clicked */
   LONG     ReleaseFrame;     /* Frame to display when the button is released */
   LONG     Thickness;        /* Border thickness */

   PRIVATE_FIELDS
   UBYTE    String[120];      /* String to display inside the button */
   struct Picture *Picture;   /* Bitmap image inside the button */
   struct   RGB ColourRGB;    /* Colour to use as the background */
   struct   RGB HighlightRGB; /* Colour to use for border highlighting */
   struct   RGB ShadowRGB;    /* Colour to use for border shadowing */
   OBJECTPTR TemplateXML;
   OBJECTID PictureID;
   OBJECTID FontID;
   MEMORYID ActionScriptMID;
   STRING   ActionScript;     /* Simple DML statement indicating what to do on button-press, e.g. "<set object="[desktop]" field="">"... */
   BYTE     KeyboardChannel;
   BYTE     State;
   BYTE     Clicked;          /* Set to TRUE if the button has been clicked.  Reverts back to FALSE when the user releases the button */
   BYTE     Active;
   WORD     ClickX;
   WORD     ClickY;
};

#define BTF_HIDE       0x00000001 /* Hide on creation, or button is currently hidden*/
#define BTF_DISABLED   0x00000002 /* Disable on creation */
#define BTF_NOGFX      0x00000004 /* Do not generate any text or image graphics using internal routines (use the Template graphics definition) */
#define BTF_NOBKGD     0x00000008 /* Do not draw background graphics (automatically set if Template graphics are used) */
#define BTF_LOCAL      0x00000010 /* On activation, execute objects locally rather than messaging them */
#define BTF_FADEBORDER 0x00000020 /* Fade the border edges into the button colour (requires thickness > 1) */
#define BTF_SUNKEN     0x00000040 /* Reverses the highlight and shadow definitions to create a sunken effect */
#define BTF_NOFOCUS    0x00000080 /* Do not draw lines around the button to indicate that it has the focus */
#define BTF_PULSE      0x00000100 /* Respond to repeated clicks occuring when the mouse button is held down */

/*****************************************************************************
** Class definition.
*/

static struct FieldDef ButtonFlags[] = {
   { "HIDE",     BTF_HIDE },
   { "DISABLED", BTF_DISABLED },
   { "NOGFX",    BTF_NOGFX },
   { "NOBKGD",   BTF_NOBKGD },
   { "LOCAL",    BTF_LOCAL },
   { "SUNKEN",   BTF_SUNKEN },
   { "NOFOCUS",  BTF_NOFOCUS },
   { "PULSE",    BTF_PULSE },
   { NULL, NULL }
};

static ERROR GET_ActionScript(struct Button *, STRING *);
static ERROR GET_Bottom(struct Button *, LONG *);
static ERROR GET_ColourRGB(struct Button *, struct RGB *);
static ERROR GET_Disable(struct Button *, LONG *);
static ERROR GET_Height(struct Button *, struct Variable *);
static ERROR GET_HighlightRGB(struct Button *, struct RGB *);
static ERROR GET_Right(struct Button *, LONG *);
static ERROR GET_ShadowRGB(struct Button *, struct RGB *);
static ERROR GET_String(struct Button *, STRING *);
static ERROR GET_Width(struct Button *, struct Variable *);
static ERROR GET_X(struct Button *, struct Variable *);
static ERROR GET_XOffset(struct Button *, struct Variable *);
static ERROR GET_Y(struct Button *, struct Variable *);
static ERROR GET_YOffset(struct Button *, struct Variable *);

static ERROR SET_ActionScript(struct Button *, STRING);
static ERROR SET_Colour(struct Button *, STRING);
static ERROR SET_Disable(struct Button *, LONG *);
static ERROR SET_Image(struct Button *, STRING);
static ERROR SET_Height(struct Button *, struct Variable *);
static ERROR SET_Highlight(struct Button *, STRING);
static ERROR SET_Shadow(struct Button *, STRING);
static ERROR SET_String(struct Button *, STRING);
static ERROR SET_TabFocus(struct Button *, OBJECTID *);
static ERROR SET_Template(struct Button *, STRING);
static ERROR SET_Width(struct Button *, struct Variable *);
static ERROR SET_X(struct Button *, struct Variable *);
static ERROR SET_XOffset(struct Button *, struct Variable *);
static ERROR SET_Y(struct Button *, struct Variable *);
static ERROR SET_YOffset(struct Button *, struct Variable *);

static struct FieldArray ButtonFields[] = {
   { "Region",         0, FDF_OBJECTID|FDF_R,   ID_RENDER, NULL, NULL },
   { "Drawable",       0, FDF_OBJECTID|FDF_RW,  ID_RENDER, NULL, NULL },
   { "Input",          0, FDF_OBJECTID|FDF_RI,  ID_KEYBOARD, NULL ,NULL },
   { "Flags",          0, FDF_LONGFLAGS|FDF_RW, (LONG)&ButtonFlags, NULL, NULL },
   { "Font",           0, FDF_CHILD|FDF_R,      0, NULL, NULL },
   { "EnterFrame",     0, FDF_LONG|FDF_RW,      0, NULL, NULL },
   { "ExitFrame",      0, FDF_LONG|FDF_RW,      0, NULL, NULL },
   { "ClickFrame",     0, FDF_LONG|FDF_RW,      0, NULL, NULL },
   { "ReleaseFrame",   0, FDF_LONG|FDF_RW,      0, NULL, NULL },
   { "Thickness",      0, FDF_LONG|FDF_RW,      0, NULL, NULL },
   /** Virtual fields ***/
   { "ActionScript",   0, FDF_STRING|FDF_RW,   0, GET_ActionScript, SET_ActionScript },
   { "Bottom",         0, FDF_LONG|FDF_R,      0, GET_Bottom,       NULL },
   { "Disable",        0, FDF_LONG|FDF_RW,     0, GET_Disable,      SET_Disable },
   { "Colour",         0, FDF_STRING|FDF_W,    0, NULL,             SET_Colour },
   { "ColourRGB",      0, FDF_POINTER|FDF_R,   0, GET_ColourRGB,    NULL },
   { "Highlight",      0, FDF_STRING|FDF_W,    0, NULL,             SET_Highlight },
   { "HighlightRGB",   0, FDF_POINTER|FDF_R,   0, GET_HighlightRGB, NULL },
   { "Image",          0, FDF_STRING|FDF_W,    0, NULL,             SET_Image },
   { "Right",          0, FDF_LONG|FDF_R,      0, GET_Right,        NULL },
   { "Shadow",         0, FDF_STRING|FDF_W,    0, NULL,             SET_Shadow },
   { "ShadowRGB",      0, FDF_POINTER|FDF_R,   0, GET_ShadowRGB,    NULL },
   { "String",         0, FDF_STRING|FDF_RW,   0, GET_String,       SET_String },
   { "TabFocus",       0, FDF_OBJECTID|FDF_W,  ID_TABFOCUS, NULL,   SET_TabFocus },
   { "Template",       0, FDF_STRING|FDF_I,    0, NULL,             SET_Template },
   { "Text",           0, FDF_STRING|FDF_RW,   0, GET_String,       SET_String },
   /*** Variable Fields ***/
   { "Height",         0, FDF_VARIABLE|FDF_DOUBLE|FDF_PERCENTAGE|FDF_RW, 0, GET_Height,  SET_Height },
   { "Width",          0, FDF_VARIABLE|FDF_DOUBLE|FDF_PERCENTAGE|FDF_RW, 0, GET_Width,   SET_Width },
   { "X",              0, FDF_VARIABLE|FDF_DOUBLE|FDF_PERCENTAGE|FDF_RW, 0, GET_X,       SET_X },
   { "XCoord",         0, FDF_VARIABLE|FDF_DOUBLE|FDF_PERCENTAGE|FDF_RW, 0, GET_X,       SET_X },
   { "XOffset",        0, FDF_VARIABLE|FDF_DOUBLE|FDF_PERCENTAGE|FDF_RW, 0, GET_XOffset, SET_XOffset },
   { "Y",              0, FDF_VARIABLE|FDF_DOUBLE|FDF_PERCENTAGE|FDF_RW, 0, GET_Y,       SET_Y },
   { "YCoord",         0, FDF_VARIABLE|FDF_DOUBLE|FDF_PERCENTAGE|FDF_RW, 0, GET_Y,       SET_Y },
   { "YOffset",        0, FDF_VARIABLE|FDF_DOUBLE|FDF_PERCENTAGE|FDF_RW, 0, GET_YOffset, SET_YOffset },
   END_FIELD
};

/*****************************************************************************
** Actions
*/

static ERROR BUTTON_AccessObject(struct Button *, APTR);
static ERROR BUTTON_ActionNotify(struct Button *, struct acActionNotify *);
static ERROR BUTTON_Activate(struct Button *, APTR);
static ERROR BUTTON_ClosingTag(struct Button *, APTR);
static ERROR BUTTON_DataChannel(struct Button *, struct acDataChannel *);
static ERROR BUTTON_Disable(struct Button *, APTR);
static ERROR BUTTON_Enable(struct Button *, APTR);
static ERROR BUTTON_Focus(struct Button *, APTR);
static ERROR BUTTON_Free(struct Button *, APTR);
static ERROR BUTTON_Hide(struct Button *, APTR);
static ERROR BUTTON_LostFocus(struct Button *, APTR);
static ERROR BUTTON_Init(struct Button *, APTR);
static ERROR BUTTON_MoveToBack(struct Button *, APTR);
static ERROR BUTTON_MoveToFront(struct Button *, APTR);
static ERROR BUTTON_NewObject(struct Button *, APTR);
static ERROR BUTTON_Redimension(struct Button *, struct acRedimension *);
static ERROR BUTTON_ReleaseObject(struct Button *, APTR);
static ERROR BUTTON_Resize(struct Button *, struct acResize *);
static ERROR BUTTON_Show(struct Button *, APTR);
static ERROR BUTTON_UserClick(struct Button *, struct acUserClick *);
static ERROR BUTTON_UserClickRelease(struct Button *, struct acUserClickRelease *);
static ERROR BUTTON_UserClickRepeat(struct Button *, struct acUserClickRepeat *);
static ERROR BUTTON_UserMovement(struct Button *, struct acUserMovement *);

static struct ActionArray ButtonActions[] = {
   { AC_AccessObject,     BUTTON_AccessObject },
   { AC_ActionNotify,     BUTTON_ActionNotify },
   { AC_Activate,         BUTTON_Activate },
   { AC_ClosingTag,       BUTTON_ClosingTag },
   { AC_DataChannel,      BUTTON_DataChannel },
   { AC_Disable,          BUTTON_Disable },
   { AC_Enable,           BUTTON_Enable },
   { AC_Focus,            BUTTON_Focus },
   { AC_Free,             BUTTON_Free },
   { AC_Hide,             BUTTON_Hide },
   { AC_Init,             BUTTON_Init },
   { AC_LostFocus,        BUTTON_LostFocus },
   { AC_MoveToBack,       BUTTON_MoveToBack },
   { AC_MoveToFront,      BUTTON_MoveToFront },
   { AC_NewObject,        BUTTON_NewObject },
   { AC_Redimension,      BUTTON_Redimension },
   { AC_ReleaseObject,    BUTTON_ReleaseObject },
   { AC_Resize,           BUTTON_Resize },
   { AC_Show,             BUTTON_Show },
   { AC_UserClick,        BUTTON_UserClick },
   { AC_UserClickRelease, BUTTON_UserClickRelease },
   { AC_UserClickRepeat,  BUTTON_UserClickRepeat },
   { AC_UserMovement,     BUTTON_UserMovement },
   { NULL, NULL }
};

/*****************************************************************************
** Prototypes
*/

static void DrawButton(struct Button *, struct Render *, struct Bitmap *);

/*****************************************************************************
** Command: Init()
*/

static ERROR CMDInit(OBJECTPTR argModule, struct KernelBase *argKernelBase)
{
   KernelBase = argKernelBase;

   /*** Open required modules ***/

   if (CreateObject(ID_MODULE, NULL, &modStrings, NULL,
      FID_Name|TSTRING,   "strings",
      FID_Version|TFLOAT, MODVERSION_STRINGS,
      TAGEND) IS ERR_Okay) {
      if (GetField(modStrings, FID_ModBase, FT_POINTER, &StringsBase) != ERR_Okay) {
         return(ObjectError(argModule, ERH_InitModule, ERR_GetField));
      }
   }
   else return(ObjectError(argModule, ERH_InitModule, ERR_CreateObject));

   if (CreateObject(ID_MODULE, NULL, &modFileSystem, NULL,
      FID_Name|TSTRING,   "filesystem",
      FID_Version|TFLOAT, MODVERSION_FILESYSTEM,
      TAGEND) IS ERR_Okay) {
      if (GetField(modFileSystem, FID_ModBase, FT_POINTER, &FileSystemBase) != ERR_Okay) {
         return(ObjectError(argModule, ERH_InitModule, ERR_GetField));
      }
   }
   else return(ObjectError(argModule, ERH_InitModule, ERR_CreateObject));

   return(CreateObject(ID_CLASS, NULL, (OBJECTPTR *)&ButtonClass, NULL,
      FID_BaseClassID|TLONG, ID_BUTTON,
      FID_Version|TFLOAT,    VER_BUTTON,
      FID_Name|TSTRING,      "Button",
      FID_Category|TLONG,    CCF_GUI,
      FID_Flags|TLONG,       CLF_PROMOTECHILDREN,
      FID_Actions|TPTR,      ButtonActions,
      FID_Fields|TPTR,       ButtonFields,
      FID_Size|TLONG,        sizeof(struct Button),
      TAGEND));
}

/*****************************************************************************
** Command: Expunge()
*/

static ERROR CMDExpunge(void)
{
   if (ButtonClass)   { acFree(ButtonClass);   ButtonClass = NULL; }
   if (modStrings)    { acFree(modStrings);    modStrings = NULL; }
   if (modFileSystem) { acFree(modFileSystem); modFileSystem = NULL; }
   return(ERR_Okay);
}

/*****************************************************************************
** Button: AccessObject
*/

static ERROR BUTTON_AccessObject(struct Button *Self, APTR Void)
{
   if (Self->FontID) {
      if (AccessObject(Self->FontID, 0, (OBJECTPTR *)&Self->Font) != ERR_Okay) {
         return(ObjectError(Self, ERH_AccessObject, ERR_AccessObject));
      }
   }
   return(ERR_Okay);
}

/*****************************************************************************
** Button: ActionNotify
*/

static ERROR BUTTON_ActionNotify(struct Button *Self, struct acActionNotify *Args)
{
   /* NB: The reason why we pass notified actions onto our support routines is
   ** for the benefit of objects that might want to subscribe to the actions
   ** of our Button object.
   */

   if (Args->ActionID IS AC_UserClick) {
      Action(AC_UserClick, Self, Args->Args);
   }
   else if (Args->ActionID IS AC_UserClickRepeat) {
      Action(AC_UserClickRepeat, Self, Args->Args);
   }
   else if (Args->ActionID IS AC_UserClickRelease) {
      Action(AC_UserClickRelease, Self, Args->Args);
   }
   else if (Args->ActionID IS AC_UserMovement) {
      Action(AC_UserMovement, Self, Args->Args);
   }
   else if (Args->ActionID IS AC_Focus) {
      OBJECTPTR input;

      DPrintF("~Focus()","[Button:%d]", Self->Head.UniqueID);

      if (Self->KeyboardChannel IS FALSE) {
         if (AccessObject(Self->InputID, 5000, &input) IS ERR_Okay) {
            SubscribeChannel(input, Self);
            Self->KeyboardChannel = TRUE;
            ReleaseObject(input);
         }
         else ObjectError(Self, ERH_Focus, ERR_AccessObject);
      }

      DelayMsg(AC_Draw, Self->RegionID, NULL);

      StepBack();
   }
   else if (Args->ActionID IS AC_LostFocus) Action(AC_LostFocus, Self, NULL);
   else if (Args->ActionID IS AC_Disable) {
      Self->Flags |= BTF_DISABLED;

      DelayMsg(AC_Draw, Self->RegionID, NULL);
   }
   else if (Args->ActionID IS AC_Enable) {
      Self->Flags &= ~BTF_DISABLED;
      DelayMsg(AC_Draw, Self->RegionID, NULL);
   }
   else return(ERR_NoSupport);

   return(ERR_Okay);
}

/*****************************************************************************

###ACTION###
Name:  Activate
Short: Activates the button.
###END###

*****************************************************************************/

static ERROR BUTTON_Activate(struct Button *Self, APTR Void)
{
   STRING script;
   LONG count;
   WORD i;

   DPrintF("~Activate()","[Button:%d]", Self->Head.UniqueID);

   if (Self->Active IS TRUE) {
      DPrintF("@Activate:","Warning - recursion detected");
      StepBack();
      return(ERR_Failed);
   }

   Self->Active = TRUE;

   if (GET_ActionScript(Self, &Self->ActionScript) IS ERR_Okay) {
      if (CreateObject(ID_SCRIPT, NF_CHILD, &script, NULL,
         FID_Target|TLONG,     Self->RegionID,
         FID_Location|TSTRING, Self->ActionScript,
         TAGEND) IS ERR_Okay) {
         acActivate(script);
         acFree(script);
      }
   }

   if ((count = TotalChildren(Self->Head.UniqueID))) {
      struct ChildEntry list[count];
      if (ListChildren(Self->Head.UniqueID, list, &count) IS ERR_Okay) {
         if (Self->Flags & BTF_LOCAL) {
            for (i=0; i < count; i++) ActionMsg(AC_Activate, list[i].ObjectID, NULL);
         }
         else for (i=0; i < count; i++) DelayMsg(AC_Activate, list[i].ObjectID, NULL);
      }
   }

   Self->Active = FALSE;

   StepBack();
   return(ERR_Okay);
}

/*****************************************************************************
** Button: ClosingTag
*/

static ERROR BUTTON_ClosingTag(struct Button *Self, APTR Void)
{
   if (!(Self->Flags & BTF_HIDE)) Action(AC_Show, Self, NULL);
   return(ERR_Okay);
}

/*****************************************************************************
** Button: DataChannel
*/

static ERROR BUTTON_DataChannel(struct Button *Self, struct acDataChannel *Args)
{
   if (!Args) return(ObjectError(Self, ERH_DataChannel, ERR_Args));

   if ((Args->DataType IS DATA_KEYPRESS) AND (Args->Version IS 1)) {
      struct dcKeyEntry *key;
      WORD i;

      if ((key = (struct dcKeyEntry *)Args->Buffer)) {
         for (i=0; i < Args->TotalEntries; i++) {

            if (key[i].Flags & KQ_PRESSED) {
               if ((key[i].Value IS K_ENTER) OR (key[i].Value IS K_NPENTER) OR (key[i].Value IS K_SPACE)) {
                  DPrintF("~DataChannel()","[Button] Enter or Space key detected.");
                     acActivate(Self);
                  StepBack();
               }
            }
         }
      }
   }

   return(ERR_Okay);
}

/*****************************************************************************

###ACTION###
Name:  Disable
Short: Turns the button off.
###END###

*****************************************************************************/

static ERROR BUTTON_Disable(struct Button *Self, APTR Void)
{
   /* See the ActionNotify routine to see what happens
   ** when the drawable is disabled.
   */
   DPrintF("Disable()","[Button:%d]", Self->Head.UniqueID);
   acDisableID(Self->RegionID);
   return(ERR_Okay);
}

/*****************************************************************************

###ACTION###
Name:  Enable
Short: Turns the button on if it has been disabled.
###END###

*****************************************************************************/

static ERROR BUTTON_Enable(struct Button *Self, APTR Void)
{
   /* See the ActionNotify routine to see what happens
   ** when the drawable is enabled.
   */
   DPrintF("Enable()","[Button:%d]", Self->Head.UniqueID);
   acEnableID(Self->RegionID);
   return(ERR_Okay);
}

/*****************************************************************************

###ACTION###
Name:  Focus
Short: Sets the focus on the button and activates keyboard monitoring.
###END###

*****************************************************************************/

static ERROR BUTTON_Focus(struct Button *Self, APTR Void)
{
   return(acFocusID(Self->RegionID));
}

/*****************************************************************************
** Button: Free
*/

static ERROR BUTTON_Free(struct Button *Self, APTR Void)
{
   if (Self->KeyboardChannel IS TRUE) {
      OBJECTPTR input;
      if (AccessObject(Self->InputID, 5000, &input) IS ERR_Okay) {
         UnsubscribeChannel(input, Self->Head.UniqueID);
         Self->KeyboardChannel = FALSE;
         ReleaseObject(input);
      }
      else ObjectError(Self, ERH_LostFocus, ERR_AccessObject);
   }

   if (Self->ActionScriptMID) {
      if (Self->ActionScript) {
         ReleaseMemory(Self->ActionScript);
         Self->ActionScript = NULL;
      }
      FreeMemoryID(Self->ActionScriptMID);
      Self->ActionScriptMID = NULL;
   }

   if (Self->PictureID) {
      if (Self->Picture) {
         acFree(Self->Picture);
         ReleaseObject(Self->Picture);
         Self->Picture = NULL;
      }
      else ActionMsg(AC_Free, Self->PictureID, NULL);
      Self->PictureID = NULL;
   }

   if (Self->FontID) {
      if (Self->Font) {
         acFree(Self->Font);
         ReleaseObject(Self->Font);
         Self->Font = NULL;
      }
      else acFreeID(Self->FontID);
      Self->FontID = NULL;
   }

   if (Self->RegionID) {
      acFreeID(Self->RegionID);
      Self->RegionID = NULL;
   }

   if (Self->TemplateXML) {
      acFree(Self->TemplateXML);
      Self->TemplateXML = NULL;
   }

   return(ERR_Okay);
}

/*****************************************************************************

###ACTION###
Name:  Hide
Short: Removes the button from the display.
###END###

*****************************************************************************/

static ERROR BUTTON_Hide(struct Button *Self, APTR Void)
{
   acHideID(Self->RegionID);
   return(ERR_Okay);
}

/*****************************************************************************
** Button: Init()
*/

static ERROR BUTTON_Init(struct Button *Self, APTR Void)
{
   struct Render *region;
   OBJECTID owner_id;

   /*** Find our parent drawable ***/

   if (!Self->DrawableID) {
      owner_id = GetOwner(Self);
      while ((owner_id) AND (GetClassID(owner_id) != ID_RENDER)) {
         owner_id = GetOwnerID(owner_id);
      }
      if (owner_id) Self->DrawableID = owner_id;
      else return(ObjectError(Self, ERH_Init, ERR_UnsupportedOwner));
   }

   /*** Setup keyboard ***/

   if (!Self->InputID) {
      if (FastFindObject("SystemKeyboard", NULL, &Self->InputID, 1, NULL) != ERR_Okay) {
         DPrintF("@Init:","[Button] Warning: Could not find any SystemKeyboard objects.");
      }
   }

   /*** Initialise the Font ***/

   if (acInit(Self->Font) != ERR_Okay) {
      return(ObjectError(Self, ERH_Init, ERR_Init));
   }

   /*** Initialise the button region ***/

   if (AccessObject(Self->RegionID, 5000, (OBJECTPTR *)&region) IS ERR_Okay) {
      SetFields(region, FID_Parent|TLONG, Self->DrawableID,
                        FID_Region|TLONG, TRUE,
                        TAGEND);

      if (!(region->Dimensions & (DMF_FIXED_HEIGHT|DMF_RELATIVE_HEIGHT))) {
         if ((!(region->Dimensions & (DMF_FIXED_Y|DMF_RELATIVE_Y))) OR (!(region->Dimensions & (DMF_FIXED_YOFFSET|DMF_RELATIVE_YOFFSET)))) {
            SetFields(region, FID_Height|TLONG, Self->Font->MaxHeight + 10, TAGEND);
         }
      }

      if (!(region->Dimensions & (DMF_FIXED_WIDTH|DMF_RELATIVE_WIDTH))) {
         if ((!(region->Dimensions & (DMF_FIXED_X|DMF_RELATIVE_X))) OR (!(region->Dimensions & (DMF_FIXED_XOFFSET|DMF_RELATIVE_XOFFSET)))) {
            struct mtCalcStringWidth calc;
            calc.String = Self->String;
            calc.Length = -1;
            Action(MT_CalcStringWidth, Self->Font, &calc);
            SetFields(region, FID_Width|TLONG, calc.Width + 24, TAGEND);
         }
      }

      if (acInit(region) IS ERR_Okay) {
         SubscribeActionTags(region, Self,
            AC_Disable,
            AC_Enable,
            AC_Focus,
            AC_LostFocus,
            AC_UserClick,
            AC_UserClickRelease,
            AC_UserClickRepeat,
            AC_UserMovement,
            TAGEND);
      }
      else {
         ReleaseObject(region);
         return(ObjectError(Self, ERH_Init, ERR_Init));
      }

      ReleaseObject(region);
   }
   else return(ObjectError(Self, ERH_Init, ERR_AccessObject));

   /*** Use the base template to create the button graphics ***/

   if ((Self->TemplateXML) AND (!(Self->Flags & BTF_NOBKGD))) {
      OBJECTPTR script;
      struct mtGetXMLString getxml;
      struct XMLTag **xmllist, *tag;

      if (GetFields(Self->TemplateXML, FID_Tags|TPTR, &xmllist, TAGEND) IS ERR_Okay) {
         for (tag=xmllist[0]; tag; tag=tag->Next) {
            if (tag->Child) {
               if (StrCompare("graphics", tag->Attributes[0].Name, 0, STR_MATCHLENGTH) IS ERR_Okay) {
                  getxml.Index = tag->Child->Index;
                  getxml.Flags = XMF_INCLUDESIBLINGS;
                  Action(MT_GetXMLString, Self->TemplateXML, &getxml);

                  if (CreateObject(ID_SCRIPT, NF_CHILD, &script, NULL,
                     FID_String|TSTRING, getxml.Result,
                     FID_Target|TLONG,   Self->RegionID,
                     TAGEND) IS ERR_Okay) {

                     if (acActivate(script) IS ERR_Okay) {
                        Self->Flags |= BTF_NOBKGD;
                     }

                     acFree(script);
                  }

                  FreeMemory(getxml.Result);
                  break;
               }
            }
         }
      }

      acFree(Self->TemplateXML);
      Self->TemplateXML = NULL;
   }

   if (AccessObject(Self->RegionID, 5000, (OBJECTPTR *)&region) IS ERR_Okay) {
      ActionTags(MT_SetRenderCallback, region, Self, &DrawButton);
      ReleaseObject(region);
   }
   else return(ObjectError(Self, ERH_Init, ERR_AccessObject));

   return(ERR_Okay);
}

/*****************************************************************************

###ACTION###
Name:  LostFocus
Short: Removes the focus from the button and disables key monitoring.
###END###

*****************************************************************************/

static ERROR BUTTON_LostFocus(struct Button *Self, APTR Void)
{
   OBJECTPTR input;

   if (Self->KeyboardChannel IS TRUE) {
      if (AccessObject(Self->InputID, 5000, &input) IS ERR_Okay) {
         UnsubscribeChannel(input, Self->Head.UniqueID);
         Self->KeyboardChannel = FALSE;
         ReleaseObject(input);
      }
      else ObjectError(Self, ERH_LostFocus, ERR_AccessObject);
   }

   DelayMsg(AC_Draw, Self->RegionID, NULL);

   return(ERR_Okay);
}

/*****************************************************************************

###ACTION###
Name:  MoveToBack
Short: Moves the button to the back of the display area.
###END###

*****************************************************************************/

static ERROR BUTTON_MoveToBack(struct Button *Self, APTR Void)
{
   ActionMsg(AC_MoveToBack, Self->RegionID, NULL);
   return(ERR_Okay);
}

/*****************************************************************************

###ACTION###
Name:  MoveToFront
Short: Moves the button to the front of the display area.
###END###

*****************************************************************************/

static ERROR BUTTON_MoveToFront(struct Button *Self, APTR Void)
{
   ActionMsg(AC_MoveToFront, Self->RegionID, NULL);
   return(ERR_Okay);
}

/*****************************************************************************
** Button: NewObject()
*/

static ERROR BUTTON_NewObject(struct Button *Self, APTR Void)
{
   if (NewObject(ID_RENDER, NF_CHILD|Self->Head.ObjectFlags, NULL, &Self->RegionID) IS ERR_Okay) {
      if (NewObject(ID_FONT, NF_CHILD|Self->Head.ObjectFlags, &Self->Font, &Self->FontID) IS ERR_Okay) {
         SetField(Self->Font, FID_Face, FT_STRING, "sans serif");

         Self->ExitFrame = 1;
         Self->ReleaseFrame = 1;

         /*** Shadow colour ***/
         Self->ShadowRGB.Red   = 100;
         Self->ShadowRGB.Green = 100;
         Self->ShadowRGB.Blue  = 100;

         /*** Internal colour ***/
         Self->ColourRGB.Red   = 210;
         Self->ColourRGB.Green = 210;
         Self->ColourRGB.Blue  = 210;

         /*** Highlight colour ***/
         Self->HighlightRGB.Red   = 255;
         Self->HighlightRGB.Green = 255;
         Self->HighlightRGB.Blue  = 255;

         Self->Thickness = 1;

         SET_Template(Self, "templates:buttondef.xml");

         return(ERR_Okay);
      }
      else return(ERR_NewObject);
   }
   else return(ERR_NewObject);
}

/*****************************************************************************

###ACTION###
Name:  Redimension
Short: Changes the size and position of the button.
###END###

*****************************************************************************/

static ERROR BUTTON_Redimension(struct Button *Self, struct acRedimension *Args)
{
   return(ActionMsg(AC_Redimension, Self->RegionID, Args));
}

/*****************************************************************************
** Button: ReleaseObject
*/

static ERROR BUTTON_ReleaseObject(struct Button *Self, APTR Void)
{
   if (Self->Font) {
      ReleaseObject(Self->Font);
      Self->Font = NULL;
   }

   return(ERR_Okay);
}

/*****************************************************************************

###ACTION###
Name:  Resize
Short: Alters the size of the button.
###END###

*****************************************************************************/

static ERROR BUTTON_Resize(struct Button *Self, struct acResize *Args)
{
   return(ActionMsg(AC_Resize, Self->RegionID, Args));
}

/*****************************************************************************

###ACTION###
Name:  Show
Short: Puts the button on display.
###END###

*****************************************************************************/

static ERROR BUTTON_Show(struct Button *Self, APTR Void)
{
   ActionMsg(AC_Show, Self->RegionID, NULL);
   return(ERR_Okay);
}

/*****************************************************************************

###ACTION###
Name:  UserClick
Short: Emulates a user click on the button.
###END###

*****************************************************************************/

static ERROR BUTTON_UserClick(struct Button *Self, struct acUserClick *Args)
{
   OBJECTPTR drawable;

   if (!Args) return(ObjectError(Self, ERH_UserClick, ERR_Args));

   if (Self->Flags & BTF_DISABLED) return(ERR_Okay|ERF_Notified);

   if (Args->ButtonFlags & JD_LMB) {
      Self->Clicked = TRUE;
      Self->ClickX  = Args->XCoord;
      Self->ClickY  = Args->YCoord;

      if (Self->ClickFrame) {
         if (AccessObject(Self->RegionID, 3000, &drawable) IS ERR_Okay) {
            SetField(drawable, FID_Frame, FT_LONG, &Self->ClickFrame);
            ReleaseObject(drawable);
         }
      }

      DelayMsg(AC_Draw, Self->RegionID, NULL);
   }

   return(ERR_Okay);
}

/*****************************************************************************

###ACTION###
Name:  UserClickRelease
Short: Emulates a user click on the button.
###END###

*****************************************************************************/

static ERROR BUTTON_UserClickRelease(struct Button *Self, struct acUserClickRelease *Args)
{
   OBJECTPTR drawable;
   WORD clickx, clicky;

   if (!Args) return(ObjectError(Self, ERH_UserClickRelease, ERR_Args));

   if ((Args->ButtonFlags & JD_LMB) AND (Self->Clicked IS TRUE)) {
      Self->Clicked = FALSE;
      clickx = Args->XCoord - Self->ClickX;
      clicky = Args->YCoord - Self->ClickY;
      if (clickx < 0) clickx = -clickx;
      if (clicky < 0) clicky = -clicky;

      /*** Change the drawable object's graphical frame if necessary ***/

      if ((Self->ClickFrame) OR (Self->ReleaseFrame)) {
         if (AccessObject(Self->RegionID, 4000, &drawable) IS ERR_Okay) {
            SetField(drawable, FID_Frame, FT_LONG, &Self->ReleaseFrame);
            ReleaseObject(drawable);
         }
      }

      ActionMsg(AC_Draw, Self->RegionID, NULL);

      if (((clickx < 4) AND (clicky < 4)) OR (Self->Flags & BTF_PULSE)) acActivate(Self);
   }

   return(ERR_Okay);
}

/*****************************************************************************

###ACTION###
Name:  UserClickRepeat
Short: Emulates a user click-repeat on the button.
###END###

*****************************************************************************/

static ERROR BUTTON_UserClickRepeat(struct Button *Self, struct acUserClickRepeat *Args)
{
   if ((Self->Flags & BTF_PULSE) AND (Args->ButtonFlags & JD_LMB)) {
      acActivate(Self);
   }

   return(ERR_Okay);
}

/*****************************************************************************

###ACTION###
Name:  UserMovement
Short: Emulates user movement over the button.
###END###

*****************************************************************************/

static ERROR BUTTON_UserMovement(struct Button *Self, struct acUserMovement *Args)
{
   struct Render *drawable;

   if (!Args) return(ObjectError(Self, ERH_UserMovement, ERR_Args));

   if (Args->OverObject IS Self->RegionID) {
      if (Self->State IS STATE_ENTERED) Self->State = STATE_INSIDE;
      else if (Self->State != STATE_INSIDE) Self->State = STATE_ENTERED;
   }
   else {
      if (Self->State IS STATE_EXITED) return(ERR_Okay);
      else Self->State = STATE_EXITED;
   }

   /*** Change the drawable's frame if necessary ***/

   if (Self->State != STATE_INSIDE) {
      if (AccessObject(Self->RegionID, 2000, (OBJECTPTR *)&drawable) IS ERR_Okay) {
         if (!(drawable->Flags & RNF_DISABLED)) {
            if (Self->State IS STATE_EXITED) {
               SetField(drawable, FID_Frame, FT_LONG, &Self->ExitFrame);
               DelayMsg(AC_Draw, Self->RegionID, NULL);
            }
            else if (Self->State IS STATE_ENTERED) {
               if (!(Self->Flags & BTF_DISABLED)) {
                  SetField(drawable, FID_Frame, FT_LONG, &Self->EnterFrame);
                  DelayMsg(AC_Draw, Self->RegionID, NULL);
               }
            }
         }
         ReleaseObject(drawable);
      }
   }

   if (Self->State IS STATE_ENTERED) Self->State = STATE_INSIDE;

   return(ERR_Okay);
}

/*****************************************************************************

###FIELD###
Name:   ActionScript
Short:  A script to be executed when the button is clicked.
Type:   STRING
Status: Get/Set

###DESCRIPTION###
<p>When a button is clicked, you can elect to run a script that performs a
responsive action.  The ActionScript string will typically refer to a file
location, although the special STRING: assignment followed with script
execution text (e.g. DML) can also be useful for optimising execution for
simple scripts.</p>

###END###

*****************************************************************************/

static ERROR GET_ActionScript(struct Button *Self, STRING *Value)
{
   if (Self->ActionScript) {
      *Value = Self->ActionScript;
      return(ERR_Okay);
   }
   else if (!Self->ActionScriptMID) {
      *Value = NULL;
      return(ERR_FieldNotSet);
   }
   else if (AccessMemory(Self->ActionScriptMID, MEM_READ, (void **)&Self->ActionScript) IS ERR_Okay) {
      *Value = Self->ActionScript;
      return(ERR_Okay);
   }
   else {
      *Value = NULL;
      return(ObjectError((OBJECTPTR)Self, ERH_GetField, ERR_AccessMemory));
   }
}

static ERROR SET_ActionScript(struct Button *Self, STRING Value)
{
   LONG i;

   if (Self->ActionScript)    { ReleaseMemory(Self->ActionScript);   Self->ActionScript = NULL; }
   if (Self->ActionScriptMID) { FreeMemoryID(Self->ActionScriptMID); Self->ActionScriptMID = NULL; }

   if ((Value) AND (*Value)) {
      for (i=0; Value[i]; i++);
      if (AllocMemory(i+1, MEM_STRING|MEM_NOCLEAR|Self->Head.MemFlags, (void **)&Self->ActionScript, &Self->ActionScriptMID) IS ERR_Okay) {
         for (i=0; Value[i]; i++) Self->ActionScript[i] = Value[i];
         Self->ActionScript[i] = 0;
      }
      else return(ObjectError((OBJECTPTR)Self, ERH_SetField, ERR_AllocMemory));
   }
   return(ERR_Okay);
}

/*****************************************************************************

###FIELD###
Name:   Bottom
Short:  The bottom coordinate of the button.
Type:   LONG
Status: Get

###DESCRIPTION###
<p>The bottom coordinate of the button (calculated as YCoord + Height) is
readable from this field.</p>

###END###

*****************************************************************************/

static ERROR GET_Bottom(struct Button *Self, LONG *Value)
{
   OBJECTPTR drawable;

   if (AccessObject(Self->RegionID, 4000, &drawable) IS ERR_Okay) {
      GetField(drawable, FID_Bottom, FT_LONG, Value);
      ReleaseObject(drawable);
      return(ERR_Okay);
   }
   else return(ObjectError(Self, ERH_GetField, ERR_AccessObject));
}

/*****************************************************************************

###FIELD###
Name:   Colour
Short:  String-based field for setting the colour inside of the button.
Type:   STRING
Status: Set

###DESCRIPTION###
<p>The colour inside of a button is declared by writing to this field.
The colour must be in hexadecimal or separated-decimal format - for example
to create a pure red colour, a setting of "#ff0000" or "255,0,0" would be
valid.</p>

<p>If you create a button without setting the Colour field, the button will
be drawn in black.</p>

###SEE ALSO###
Fields: Highlight, Shadow, Thickness

###END###

*****************************************************************************/

static ERROR SET_Colour(struct Button *Self, STRING Colour)
{
   if (Colour) {
      StrReadColour(Colour, &Self->ColourRGB.Red, &Self->ColourRGB.Green, &Self->ColourRGB.Blue, &Self->ColourRGB.Alpha);
   }
   return(ERR_Okay);
}

/*****************************************************************************

###FIELD###
Name:   ClickFrame
Short:  The graphics frame to display when the button is clicked.
Type:   LONG
Status: Read/Write

###DESCRIPTION###
<p>This field specifies the drawable frame to switch to when the user
clicks on the button.  By default this field is initially set to
zero, which has no effect on the drawable frame.  When the user releases
the button, it will revert to the frame indicated by the <a href="#ReleaseFrame">ReleaseFrame</a>
field.</p>

###SEE ALSO###
Field: ReleaseFrame

###END###

*****************************************************************************/

/*****************************************************************************

###FIELD###
Name:   ColourRGB
Short:  Defines the internal colour of the button in RGB format.
Type:   RGB
Status: Read

###DESCRIPTION###
<p>The colour inside of the button can be read in RGB format from this field.</p>

###SEE ALSO###
Field: Colour

###END###

*****************************************************************************/

static ERROR GET_ColourRGB(struct Button *Self, struct RGB *Value)
{
   Value->Red   = Self->ColourRGB.Red;
   Value->Green = Self->ColourRGB.Green;
   Value->Blue  = Self->ColourRGB.Blue;
   Value->Alpha = Self->ColourRGB.Alpha;
   return(ERR_Okay);
}

static ERROR GET_Disable(struct Button *Self, LONG *Value)
{
   if (Self->Flags & BTF_DISABLED) *Value = TRUE;
   else *Value = FALSE;
   return(ERR_Okay);
}

static ERROR SET_Disable(struct Button *Self, LONG *Value)
{
   if (*Value IS TRUE) acDisable(Self);
   else if (*Value IS FALSE) acEnable(Self);
   return(ERR_Okay);
}

/*****************************************************************************

###FIELD###
Name:   Drawable
Short:  The drawable that will contain the button graphic.
Type:   OBJECTID
Status: Read/Init

###DESCRIPTION###
<p>The drawable that will contain the button graphic is set here.  If this
field is not set prior to initialisation, the button will attempt to scan
for the correct drawable by analysing its parents until it finds a suitable
candidate.</p>

###END###

*****************************************************************************/

/*****************************************************************************

###FIELD###
Name:   EnterFrame
Short:  The graphics frame to display when the user's pointer enters the button area.
Type:   LONG
Status: Read/Write

###DESCRIPTION###
<p>This field specifies the drawable frame to switch to when the user's
pointer enters the button area.  By default this field is initially set to
zero, which has no effect on the drawable frame.</p>

###SEE ALSO###
Field: ExitFrame

###END###

*****************************************************************************/

/*****************************************************************************

###FIELD###
Name:   ExitFrame
Short:  The graphics frame to display when the user's pointer leaves the button area.
Type:   LONG
Status: Read/Write

###DESCRIPTION###
<p>This field specifies the drawable frame to switch to when the user's pointer
leaves the button area.  This field is unused if the EnterFrame field has
not been set.</p>

###SEE ALSO###
Field: EnterFrame

###END###

*****************************************************************************/

/*****************************************************************************

###FIELD###
Name:   Flags
Short:  Optional flags may be set here.
Type:   LONG
Status: Read/Init

###DESCRIPTION###
<p>Optional flags that may be set against a button object are as follows:</p>

<table border="1" bordercolordark="#000000" cellspacing="0" cellpadding="2" bordercolorlight="#000000" width="80%" align="center">
<tr><td class="tableheader" bgcolor="#eeeeee">Flag</td><td class="tableheader" bgcolor="#eeeeee">Description</td></tr>
<tr><td width="10%">HIDE</td><td>This flag may be set on initialisation in order to hide the button on its creation (applies to script usage only).</td></tr>
<tr><td width="10%">DISABLED</td><td>This readable flag is automatically set when the button is disabled.  You may set it on initialisation if the button is to be disabled on creation.</td></tr>
<tr><td width="10%">LOCAL</td><td>When the button is activated, any child objects belonging to the button will be executed immediately within the button's execution process, rather than through the messaging system.</td></tr>
<tr><td width="10%">NOBKGD</td><td>Turns off the button's background graphics (border and background colour).</td></tr>
<tr><td width="10%">NOGFX</td><td>Turns off the button's foreground graphics (text and images).</td></tr>
<tr><td width="10%">NOFOCUS</td><td>Turns off the lines that are drawn around the button border when it has the user focus.</td></tr>
</table>

###END###

*****************************************************************************/

/*****************************************************************************

###FIELD###
Name:   Height
Short:  Defines the height of a button.
Type:   DOUBLE/PERCENTAGE
Status: Get/Set

###DESCRIPTION###
<p>A button can be given a fixed or relative height by setting this field
to the desired value.  To set a relative height, use the FD_PERCENT flag when
setting the field.</p>

###END###

*****************************************************************************/

static ERROR GET_Height(struct Button *Self, struct Variable *Value)
{
   OBJECTPTR drawable;
   DOUBLE value;

   if (AccessObject(Self->RegionID, 4000, &drawable) IS ERR_Okay) {
      GetField(drawable, FID_Height, FT_DOUBLE, &value);
      ReleaseObject(drawable);

      if (Value->Type & FD_DOUBLE) Value->Double = value;
      else if (Value->Type & FD_LARGE) Value->Large = value;
      return(ERR_Okay);
   }
   else return(ObjectError(Self, ERH_GetHeight, ERR_AccessObject));
}

static ERROR SET_Height(struct Button *Self, struct Variable *Value)
{
   OBJECTPTR drawable;

   if (((Value->Type & FD_DOUBLE) AND (!Value->Double)) OR
       ((Value->Type & FD_LARGE) AND (!Value->Large))) {
      return(ERR_Okay);
   }

   if (AccessObject(Self->RegionID, 4000, &drawable) IS ERR_Okay) {
      SetField(drawable, FID_Height, FT_VARIABLE, Value);
      ReleaseObject(drawable);
      return(ERR_Okay);
   }
   else return(ERR_AccessObject);
}

/*****************************************************************************

###FIELD###
Name:   Highlight
Short:  String-based field for setting the button highlight.
Type:   STRING
Status: Set

###DESCRIPTION###
<p>The border highlight of a button is set by writing to this field. The colour
must be in hexadecimal or separated-decimal format - for example to create a
pure red colour, a setting of "#ff0000" or "255,0,0" would be valid.</p>

###SEE ALSO###
Fields: Colour, Shadow, Thickness

###END###

*****************************************************************************/

static ERROR SET_Highlight(struct Button *Self, STRING Value)
{
   if (Value) {
      StrReadColour(Value, &Self->HighlightRGB.Red, &Self->HighlightRGB.Green, &Self->HighlightRGB.Blue, &Self->HighlightRGB.Alpha);
   }
   return(ERR_Okay);
}

/*****************************************************************************

###FIELD###
Name:   HighlightRGB
Short:  Defines border highlight of the button, in RGB format.
Type:   RGB
Status: Read

###DESCRIPTION###
<p>The border highlight of the button can be read in RGB format from this field.</p>

###SEE ALSO###
Field: Colour

###END###

*****************************************************************************/

static ERROR GET_HighlightRGB(struct Button *Self, struct RGB *Value)
{
   Value->Red   = Self->HighlightRGB.Red;
   Value->Green = Self->HighlightRGB.Green;
   Value->Blue  = Self->HighlightRGB.Blue;
   Value->Alpha = Self->HighlightRGB.Alpha;
   return(ERR_Okay);
}

/*****************************************************************************

###FIELD###
Name:   Image
Short:  The image field can be set in order to load a bitmap into the button.
Type:   STRING
Status: Set

###DESCRIPTION###
<p>To display an image inside the button, set the Image field.  The image will
be displayed on the left side of the text inside the button.  If no text string
has been set, the image will be shown at the exact center of the button.</p>

###END###

*****************************************************************************/

static ERROR SET_Image(struct Button *Self, STRING Value)
{
   if (CreateObject(ID_PICTURE, NF_CHILD, &Self->Picture, NULL,
      FID_Location|TSTRING, Value,
      TAGEND) IS ERR_Okay) {
      Action(AC_Activate, Self->Picture, NULL);
      return(ERR_Okay);
   }
   else return(ERR_CreateObject);
}

/*****************************************************************************

###FIELD###
Name:   Region
Short:  The drawable that represents the button is referenced through this field.
Type:   OBJECTID
Status: Read

###DESCRIPTION###
<p>The rendering area that represents the button display can be accessed
through this field.  For further information, refer to the Render class.  Note
that talking to the drawable directly can have adverse effects on the button
control system.  Where possible, all communication should be limited to the
button object itself.</p>

###SEE ALSO###
Class: Render

###END###

*****************************************************************************/

/*****************************************************************************

###FIELD###
Name:   ReleaseFrame
Short:  The graphics frame to display when a user-click is released.
Type:   LONG
Status: Read/Write

###DESCRIPTION###
<p>If the <a href="#ClickFrame">ClickFrame</a> field has been set, you may want to match that value
by indicating the frame that should be used when the click is released.
By default, the value in this field will initially be set to 1.  This field
is unused if the <a href="#ClickFrame">ClickFrame</a> field has not been set.</p>

###END###

*****************************************************************************/

/*****************************************************************************

###FIELD###
Name:   Right
Short:  The right coordinate of the button.
Type:   LONG
Status: Get

###DESCRIPTION###
<p>The right coordinate of the button (calculated as XCoord + Width) is
readable from this field.</p>

###END###

*****************************************************************************/

static ERROR GET_Right(struct Button *Self, LONG *Value)
{
   OBJECTPTR drawable;

   if (AccessObject(Self->RegionID, 4000, &drawable) IS ERR_Okay) {
      GetField(drawable, FID_Right, FT_LONG, Value);
      ReleaseObject(drawable);
      return(ERR_Okay);
   }
   else return(ObjectError(Self, ERH_GetField, ERR_AccessObject));
}

/*****************************************************************************

###FIELD###
Name:   Shadow
Short:  String-based field for setting the button shadow.
Type:   STRING
Status: Set

###DESCRIPTION###
<p>The border shadow of a button is set by writing to this field. The colour
must be in hexadecimal or separated-decimal format - for example to create a
pure red colour, a setting of "#ff0000" or "255,0,0" would be valid.</p>

###SEE ALSO###
Fields: Colour, Highlight, Thickness

###END###

*****************************************************************************/

static ERROR SET_Shadow(struct Button *Self, STRING Value)
{
   if (Value) {
      StrReadColour(Value, &Self->ShadowRGB.Red, &Self->ShadowRGB.Green, &Self->ShadowRGB.Blue, &Self->ShadowRGB.Alpha);
   }
   return(ERR_Okay);
}

/*****************************************************************************

###FIELD###
Name:   ShadowRGB
Short:  Defines border shadow of the button, in RGB format.
Type:   RGB
Status: Read

###DESCRIPTION###
<p>The border shadow of the button can be read in RGB format from this field.</p>

###SEE ALSO###
Field: Colour

###END###

*****************************************************************************/

static ERROR GET_ShadowRGB(struct Button *Self, struct RGB *Value)
{
   Value->Red   = Self->ShadowRGB.Red;
   Value->Green = Self->ShadowRGB.Green;
   Value->Blue  = Self->ShadowRGB.Blue;
   Value->Alpha = Self->ShadowRGB.Alpha;
   return(ERR_Okay);
}

/*****************************************************************************

###FIELD###
Name:   String
Short:  The string that is to be printed inside the button is declared here.
Type:   STRING
Status: Get/Set

###DESCRIPTION###
<p>The string that you would like to be displayed in the button is
specified in this field.  The string must be in UTF-8 format and be no
longer than one line.</p>

<p>If the string is changed after initialisation, the button will be redrawn
to show the updated text.</p>

###END###

*****************************************************************************/

static ERROR GET_String(struct Button *Self, STRING *Value)
{
   if (Self->String) {
      *Value = Self->String;
      return(ERR_Okay);
   }
   else return(ERR_FieldNotSet);
}

static ERROR SET_String(struct Button *Self, STRING Value)
{
   WORD i;

   if (Value) {
      for (i=0; (i < sizeof(Self->String)-1) AND (Value[i]); i++) Self->String[i] = Value[i];
      Self->String[i] = 0;
   }
   else Self->String[0] = 0;

   /*** Send a redraw message ***/

   if (Self->Head.ObjectFlags & NF_INITIALISED) DelayMsg(AC_Draw, Self->RegionID, NULL);

   return(ERR_Okay);
}

/*****************************************************************************

###FIELD###
Name:   TabFocus
Short:  Set this field to a TabFocus object to register the button in a tab-list.
Type:   OBJECTID
Status: Set

###DESCRIPTION###
<p>The TabFocus field provides a convenient way of adding the button to a
TabFocus object, so that it can be focussed on via the tab key.  Simply set
this field to the ID of the TabFocus object that is managing the tab-list
for the application window.</p>

###SEE ALSO###
Class: TabFocus

###END###

*****************************************************************************/

static ERROR SET_TabFocus(struct Button *Self, OBJECTID *Value)
{
   OBJECTPTR tabfocus;

   if (AccessObject(*Value, 5000, &tabfocus) IS ERR_Okay) {
      if (tabfocus->ClassID IS ID_TABFOCUS) {
         ActionTags(MT_AddTabObject, tabfocus, Self->RegionID);
      }
      ReleaseObject(tabfocus);
   }
   else return(ERR_AccessObject);

   return(ERR_Okay);
}

/*****************************************************************************

###FIELD###
Name:   Template
Short:  Defines the makeup of the button using a pre-defined template.
Type:   STRING
Status: Init

###DESCRIPTION###
<p>Setting the Template field allows a button defintion to be loaded from a
pre-defined template file.  The template may contain settings for all of the
field values of a button object, as well as a customised graphics script
for building the look and feel of the button.  The template must be provided
as an XML file that can contain two possible root tags - 'value' and 'graphics'.
The value tag contains a list of field values for setting against the button,
while the graphics tag contains the graphical makeup of the button.  The
following example illustrates how a customised button can be created:</p>

<pre>
&lt;?xml version="1.0"?&gt;

&lt;values&gt;
  &lt;exitframe value="1"/&gt;
  &lt;clickframe value="2"/&gt;
  &lt;enterframe value="3"/&gt;
  &lt;releaseframe value="1"/&gt;
&lt;/values&gt;

&lt;graphics&gt;
  &lt;!-- Normal --&gt;
  &lt;box frame="1" boxes="(r0,0,!0,!0) (r1,1,!1,!1)" colour="230,230,230"
    highlight(0)="120,120,120" highlight(1)="255,255,255" shadow(0)="80,80,80"
    shadow(1)="200,200,200"/&gt;

  &lt;!-- Clicked --&gt;
  &lt;box frame="2" boxes="(s0,0,!0,!0)" colour="200,200,200"
    highlight="120,120,120" shadow="80,80,80"/&gt;

  &lt;!-- OnMovement --&gt;
  &lt;box frame="3" boxes="(r0,0,!0,!0) (r1,1,!1,!1)" colour="245,245,245"
    highlight="120,120,120" highlight(1)="255,255,255" shadow="80,80,80"
    shadow(1)="200,200,200"/&gt;
&lt;/graphics&gt;
</pre>

###END###

*****************************************************************************/

static ERROR SET_Template(struct Button *Self, STRING Value)
{
   struct XMLTag **xmllist, *tag, *child;
   UBYTE value[300];
   WORD i;

   if (Self->TemplateXML) {
      acFree(Self->TemplateXML);
      Self->TemplateXML = NULL;
   }

   if (AnalyseLocation(Value, NULL) IS ERR_Okay) {
      if (CreateObject(ID_XML, NF_CHILD, &Self->TemplateXML, NULL,
         FID_Location|TSTRING, Value,
         TAGEND) IS ERR_Okay) {

         if (GetFields(Self->TemplateXML, FID_Tags|TPTR, &xmllist, TAGEND) IS ERR_Okay) {
            for (tag=xmllist[0]; tag; tag=tag->Next) {
               if (tag->Child) {
                  if (StrCompare("values", tag->Attributes[0].Name, 0, STR_MATCHLENGTH) IS ERR_Okay) {
                     for (child=tag->Child; child; child=child->Next) {
                        if (child->TotalAttributes > 1) {
                           for (i=0; (i < sizeof(value)-1) AND (child->Attributes[1].Value[i]); i++) value[i] = child->Attributes[1].Value[i];
                           value[i] = 0;
                           StrTranslate(value, sizeof(value), NULL);
                           SetFieldVariable(Self, child->Attributes[0].Name, value);
                        }
                     }
                     break;
                  }
               }
            }
         }
      }
   }

   return(ERR_Okay);
}

/*****************************************************************************

###FIELD###
Name:   Thickness
Short:  The thickness of the button border.
Type:   LONG
Status: Read/Write

###DESCRIPTION###
<p>This field specifies the thickness of the button border, expressed in
pixels.  By default the thickness is set to a value of 1.</p>

###END###

*****************************************************************************/

/*****************************************************************************

###FIELD###
Name:   Width
Short:  Defines the width of a button.
Type:   DOUBLE/PERCENTAGE
Status: Get/Set

###DESCRIPTION###
<p>A button can be given a fixed or relative width by setting this field
to the desired value.  To set a relative width, use the FD_PERCENT flag when
setting the field.</p>

###END###

*****************************************************************************/

static ERROR GET_Width(struct Button *Self, struct Variable *Value)
{
   OBJECTPTR drawable;
   DOUBLE value;

   if (AccessObject(Self->RegionID, 4000, &drawable) IS ERR_Okay) {
      GetField(drawable, FID_Width, FT_DOUBLE, &value);
      ReleaseObject(drawable);

      if (Value->Type & FD_DOUBLE) Value->Double = value;
      else if (Value->Type & FD_LARGE) Value->Large = value;
      return(ERR_Okay);
   }
   else return(ObjectError(Self, ERH_GetWidth, ERR_AccessObject));
}

static ERROR SET_Width(struct Button *Self, struct Variable *Value)
{
   OBJECTPTR drawable;

   if (((Value->Type & FD_DOUBLE) AND (!Value->Double)) OR
       ((Value->Type & FD_LARGE) AND (!Value->Large))) {
      return(ERR_Okay);
   }

   if (AccessObject(Self->RegionID, 4000, &drawable) IS ERR_Okay) {
      SetField(drawable, FID_Width, FT_VARIABLE, Value);
      ReleaseObject(drawable);
      return(ERR_Okay);
   }
   else return(ERR_AccessObject);
}

/*****************************************************************************

###FIELD###
Name:   XCoord
Short:  The horizontal position of a button.
Type:   DOUBLE/PERCENTAGE
Status: Get/Set

###DESCRIPTION###
<p>The horizontal position of a button can be set to an absolute or relative
coordinate by writing a value to the XCoord field.  To set a
relative/percentage based value, you must use the FD_PERCENT flag or the
value will be interpreted as fixed.  Negative values are permitted.</p>

###SEE ALSO###
Field: XOffset, YCoord

###END###

*****************************************************************************/

static ERROR GET_X(struct Button *Self, struct Variable *Value)
{
   OBJECTPTR drawable;
   DOUBLE value;

   if (AccessObject(Self->RegionID, 4000, &drawable) IS ERR_Okay) {
      GetField(drawable, FID_XCoord, FT_DOUBLE, &value);
      ReleaseObject(drawable);

      if (Value->Type & FD_DOUBLE) Value->Double = value;
      else if (Value->Type & FD_LARGE) Value->Large = value;
      return(ERR_Okay);
   }
   else return(ObjectError(Self, ERH_GetX, ERR_AccessObject));
}

static ERROR SET_X(struct Button *Self, struct Variable *Value)
{
   OBJECTPTR drawable;

   if (AccessObject(Self->RegionID, 4000, &drawable) IS ERR_Okay) {
      SetField(drawable, FID_XCoord, FT_VARIABLE, Value);
      ReleaseObject(drawable);
      return(ERR_Okay);
   }
   else return(ERR_AccessObject);
}

/*****************************************************************************

###FIELD###
Name:   XOffset
Short:  The horizontal offset of a button.
Type:   DOUBLE/PERCENTAGE
Status: Get/Set

###DESCRIPTION###
<p>The XOffset has a dual purpose depending on whether or not it is set in
conjunction with an X coordinate or a Width based field.</p>

<p>If set in conjunction with an X coordinate then the button will be
drawn from that X coordinate up to the width of the container, minus the
value given in the XOffset.  This means that the width of the Button is
dynamically calculated in relation to the width of the container.</p>

<p>If the XOffset field is set in conjunction with a fixed or relative width
then the button will be drawn at an X coordinate calculated from the formula
&quot;XCoord = ContainerWidth - ButtonWidth - XOffset&quot;.</p>

###SEE ALSO###
Field: XCoord, Width, YOffset

###END###

*****************************************************************************/

static ERROR GET_XOffset(struct Button *Self, struct Variable *Value)
{
   OBJECTPTR drawable;
   DOUBLE value;

   if (AccessObject(Self->RegionID, 4000, &drawable) IS ERR_Okay) {
      GetField(drawable, FID_XOffset, FT_DOUBLE, &value);
      ReleaseObject(drawable);

      if (Value->Type & FD_DOUBLE) Value->Double = value;
      else if (Value->Type & FD_LARGE) Value->Large = value;
      return(ERR_Okay);
   }
   else return(ObjectError(Self, ERH_GetXOffset, ERR_AccessObject));
}

static ERROR SET_XOffset(struct Button *Self, struct Variable *Value)
{
   OBJECTPTR drawable;

   if (AccessObject(Self->RegionID, 4000, &drawable) IS ERR_Okay) {
      SetField(drawable, FID_XOffset, FT_VARIABLE, Value);
      ReleaseObject(drawable);
      return(ERR_Okay);
   }
   else return(ERR_AccessObject);
}

/*****************************************************************************

###FIELD###
Name:   YCoord
Short:  The vertical position of a button.
Type:   DOUBLE/PERCENTAGE
Status: Get/Set

###DESCRIPTION###
<p>The vertical position of a Button can be set to an absolute or relative
coordinate by writing a value to the YCoord field.  To set a
relative/percentage based value, you must use the FD_PERCENT flag or the
value will be interpreted as fixed.  Negative values are permitted.</p>

###SEE ALSO###
Field: YOffset, XCoord

###END###

*****************************************************************************/

static ERROR GET_Y(struct Button *Self, struct Variable *Value)
{
   OBJECTPTR drawable;
   DOUBLE value;

   if (AccessObject(Self->RegionID, 4000, &drawable) IS ERR_Okay) {
      GetField(drawable, FID_YCoord, FT_DOUBLE, &value);
      ReleaseObject(drawable);

      if (Value->Type & FD_DOUBLE) Value->Double = value;
      else if (Value->Type & FD_LARGE) Value->Large = value;
      return(ERR_Okay);
   }
   else return(ObjectError(Self, ERH_GetY, ERR_AccessObject));

}

static ERROR SET_Y(struct Button *Self, struct Variable *Value)
{
   OBJECTPTR drawable;

   if (AccessObject(Self->RegionID, 4000, &drawable) IS ERR_Okay) {
      SetField(drawable, FID_YCoord, FT_VARIABLE, Value);
      ReleaseObject(drawable);
      return(ERR_Okay);
   }
   else return(ERR_AccessObject);
}

/*****************************************************************************

###FIELD###
Name:   YOffset
Short:  The vertical offset of a button.
Type:   DOUBLE/PERCENTAGE
Status: Get/Set

###DESCRIPTION###
<p>The YOffset has a dual purpose depending on whether or not it is set in
conjunction with a Y coordinate or a Height based field.</p>

<p>If set in conjunction with a Y coordinate then the button will be drawn
from that Y coordinate up to the height of the container, minus the value
given in the YOffset.  This means that the height of the button is
dynamically calculated in relation to the height of the container.</p>

<p>If the YOffset field is set in conjunction with a fixed or relative height
then the button will be drawn at a Y coordinate calculated from the formula
&quot;YCoord = ContainerHeight - ButtonHeight - YOffset&quot;.</p>

###SEE ALSO###
Field: YCoord, Height, XOffset

###END###

*****************************************************************************/

static ERROR GET_YOffset(struct Button *Self, struct Variable *Value)
{
   OBJECTPTR drawable;
   DOUBLE value;

   if (AccessObject(Self->RegionID, 4000, &drawable) IS ERR_Okay) {
      GetField(drawable, FID_YOffset, FT_DOUBLE, &value);
      ReleaseObject(drawable);

      if (Value->Type & FD_DOUBLE) Value->Double = value;
      else if (Value->Type & FD_LARGE) Value->Large = value;
      return(ERR_Okay);
   }
   else return(ObjectError(Self, ERH_GetYOffset, ERR_AccessObject));
}

static ERROR SET_YOffset(struct Button *Self, struct Variable *Value)
{
   OBJECTPTR drawable;

   if (AccessObject(Self->RegionID, 4000, &drawable) IS ERR_Okay) {
      SetField(drawable, FID_YOffset, FT_VARIABLE, Value);
      ReleaseObject(drawable);
      return(ERR_Okay);
   }
   else return(ERR_AccessObject);
}

/*****************************************************************************
** Internal: DrawButton()
*/

static void DrawButton(struct Button *Self, struct Render *Drawable, struct Bitmap *Bitmap)
{
   struct mtDrawRGBRectangle rect;
   struct Font *font;
   ULONG highlight, shadow, temp;
   WORD i, x, y, skip;

   font = Self->Font;

   SetField(font, FID_Bitmap, FT_POINTER, Bitmap);

   if (!(Self->Flags & BTF_NOBKGD)) {
      rect.XCoord = 0;
      rect.YCoord = 0;
      rect.Width  = Drawable->Width;
      rect.Height = Drawable->Height;
      rect.Fill   = TRUE;
      rect.Red    = Self->ColourRGB.Red;
      rect.Green  = Self->ColourRGB.Green;
      rect.Blue   = Self->ColourRGB.Blue;
      if ((Self->State IS STATE_INSIDE) OR (Self->State IS STATE_ENTERED)) {
         if (!(Drawable->Flags & RNF_DISABLED)) {
            if ((rect.Red   += 20) > 255) rect.Red   = 255;
            if ((rect.Green += 20) > 255) rect.Green = 255;
            if ((rect.Blue  += 20) > 255) rect.Blue  = 255;
         }
      }
      Action(MT_DrawRGBRectangle, Bitmap, &rect);

      if (Self->Clicked) {
         highlight = PackPixel(Bitmap, Self->ShadowRGB.Red, Self->ShadowRGB.Green, Self->ShadowRGB.Blue);
         shadow    = PackPixel(Bitmap, Self->HighlightRGB.Red, Self->HighlightRGB.Green, Self->HighlightRGB.Blue);
      }
      else {
         shadow    = PackPixel(Bitmap, Self->ShadowRGB.Red, Self->ShadowRGB.Green, Self->ShadowRGB.Blue);
         highlight = PackPixel(Bitmap, Self->HighlightRGB.Red, Self->HighlightRGB.Green, Self->HighlightRGB.Blue);
      }

      if (Self->Flags & BTF_SUNKEN) {
         /*** Reverse the border definitions in sunken mode ***/
         temp      = highlight;
         highlight = shadow;
         shadow    = temp;
      }

      if (Self->Thickness > 0) {
         for (i=0; i < Self->Thickness; i++) {
            /*** Top, Bottom ***/

            ActionTags(MT_DrawRectangle, Bitmap, i, i, Drawable->Width-i-i, 1, highlight, TRUE);
            ActionTags(MT_DrawRectangle, Bitmap, i, Drawable->Height-i-1, Drawable->Width-i-i, 1, shadow, TRUE);

            /*** Left, Right ***/

            ActionTags(MT_DrawRectangle, Bitmap, i,                   i+1, 1, Drawable->Height-i-i-2, highlight, TRUE);
            ActionTags(MT_DrawRectangle, Bitmap, Drawable->Width-i-1, i+1, 1, Drawable->Height-i-i-2, shadow, TRUE);
         }
      }
   }

   SetField(font, FID_String, FT_STRING, Self->String);

   if (Drawable->Flags & RNF_DISABLED) SetFields(font, FID_Opacity|TLONG, 25, TAGEND);

   font->XCoord = 0;
   font->YCoord = 0;
   font->AlignFlags |= ALIGN_CENTER;
   font->AlignWidth = Drawable->Width;
   font->AlignHeight = Drawable->Height;
   Action(AC_Draw, font, NULL);

   if (Drawable->Flags & RNF_DISABLED) SetFields(font, FID_Opacity|TLONG, 100, TAGEND);

   if ((Drawable->Flags & RNF_HASFOCUS) AND (!(Self->Flags & BTF_NOFOCUS))) {
      if (!(Drawable->Flags & RNF_DISABLED)) {
         /*** Draw a dotted region around the button to show that it has the user focus ***/

         skip=1;
         for (x=Self->Thickness+1; x < Drawable->Width - 2; x++) {
            if (skip++ & 1) {
               Bitmap->DrawUCPixel(Bitmap, Bitmap->XOffset + x, Bitmap->YOffset + Self->Thickness + 1, 0);
               Bitmap->DrawUCPixel(Bitmap, Bitmap->XOffset + x, Bitmap->YOffset + Drawable->Height - Self->Thickness - 2, 0);
            }
         }

         /*** Left/Right ***/

         x--;
         for (y=Self->Thickness+2; y < Drawable->Height - 2; y++) {
            if (skip++ & 1) {
               Bitmap->DrawUCPixel(Bitmap, Bitmap->XOffset + x, Bitmap->YOffset + y, 0);
               Bitmap->DrawUCPixel(Bitmap, Bitmap->XOffset + Self->Thickness + 1, Bitmap->YOffset + y, 0);
            }
         }
      }
   }
}


